import numpy as np
import scipy.stats as ss


class Xi:
    """
    x and y are the data vectors
    """

    def __init__(self, x, y):

        self.x = x
        self.y = y

    @property
    def sample_size(self):
        return len(self.x)

    @property
    def x_ordered_rank(self):
        # PI is the rank vector for x, with ties broken at random
        # Not mine: source (https://stackoverflow.com/a/47430384/1628971)
        # random shuffling of the data - reason to use random.choice is that
        # pd.sample(frac=1) uses the same randomizing algorithm
        len_x = len(self.x)
        randomized_indices = np.random.choice(np.arange(len_x), len_x, replace=False)
        randomized = [self.x[idx] for idx in randomized_indices]
        # same as pandas rank method 'first'
        rankdata = ss.rankdata(randomized, method="ordinal")
        # Reindexing based on pairs of indices before and after
        unrandomized = [
            rankdata[j] for i, j in sorted(zip(randomized_indices, range(len_x)))
        ]
        return unrandomized

    @property
    def y_rank_max(self):
        # f[i] is number of j s.t. y[j] <= y[i], divided by n.
        return ss.rankdata(self.y, method="max") / self.sample_size

    @property
    def g(self):
        # g[i] is number of j s.t. y[j] >= y[i], divided by n.
        return ss.rankdata([-i for i in self.y], method="max") / self.sample_size

    @property
    def x_ordered(self):
        # order of the x's, ties broken at random.
        return np.argsort(self.x_ordered_rank)

    @property
    def x_rank_max_ordered(self):
        # Rearrange f according to ord.
        return [self.y_rank_max[i] for i in self.x_ordered]

    @property
    def mean_absolute(self):
        return (
            np.mean(
                np.abs(
                    [
                        x - y
                        for x, y in zip(
                            self.x_rank_max_ordered[0 : (self.sample_size - 1)],
                            self.x_rank_max_ordered[1 : self.sample_size],
                        )
                    ]
                )
            )
            * (self.sample_size - 1)
            / (2 * self.sample_size)
        )

    @property
    def inverse_g_mean(self):
        return np.mean(self.g * (1 - self.g))

    @property
    def correlation(self):
        """xi correlation"""
        return 1 - self.mean_absolute / self.inverse_g_mean

    @classmethod
    def xi(cls, x, y):
        return cls(x, y)

    def pval_asymptotic(self, ties=False, nperm=1000):
        """
        Returns p values of the correlation

        Args:
            ties: boolean
                If ties is true, the algorithm assumes that the data has ties
                and employs the more elaborated theory for calculating
                the P-value. Otherwise, it uses the simpler theory. There is
                no harm in setting tiles True, even if there are no ties.
            nperm: int
                The number of permutations for the permutation test, if needed.
                default 1000

        Returns:
            p value
        """
        # If there are no ties, return xi and theoretical P-value:

        if ties:
            return 1 - ss.norm.cdf(
                np.sqrt(self.sample_size) * self.correlation / np.sqrt(2 / 5)
            )

        # If there are ties, and the theoretical method
        # is to be used for calculation P-values:
        # The following steps calculate the theoretical variance
        # in the presence of ties:
        sorted_ordered_x_rank = sorted(self.x_rank_max_ordered)

        ind = [i + 1 for i in range(self.sample_size)]
        ind2 = [2 * self.sample_size - 2 * ind[i - 1] + 1 for i in ind]

        a = (
            np.mean([i * j * j for i, j in zip(ind2, sorted_ordered_x_rank)])
            / self.sample_size
        )

        c = (
            np.mean([i * j for i, j in zip(ind2, sorted_ordered_x_rank)])
            / self.sample_size
        )

        cq = np.cumsum(sorted_ordered_x_rank)

        m = [
            (i + (self.sample_size - j) * k) / self.sample_size
            for i, j, k in zip(cq, ind, sorted_ordered_x_rank)
        ]

        b = np.mean([np.square(i) for i in m])
        v = (a - 2 * b + np.square(c)) / np.square(self.inverse_g_mean)

        return 1 - ss.norm.cdf(
            np.sqrt(self.sample_size) * self.correlation / np.sqrt(v)
        )



import random
import numpy as np
import pytest

from xicor.xicor import Xi

"""
From Wikipedia:
Anscombe's quartet comprises four data sets that have nearly
identical simple descriptive statistics, yet have very different distributions
and appear very different when graphed. Each dataset consists of eleven
(x,y) points. They were constructed in 1973 by the
statistician Francis Anscombe to demonstrate both the importance of graphing
data before analyzing it and the effect of outliers and other influential
observations on statistical properties.
"""


@pytest.fixture
def anscombes_xis(anscombes_quartet):
    random.seed(2020)
    np.random.seed(2020)
    xis = {
        "xi_1": Xi(anscombes_quartet["x_1"], anscombes_quartet["y_1"]),
        "xi_2": Xi(anscombes_quartet["x_2"], anscombes_quartet["y_2"]),
        "xi_3": Xi(anscombes_quartet["x_3"], anscombes_quartet["y_3"]),
        "xi_4": Xi(anscombes_quartet["x_4"], anscombes_quartet["y_4"]),
    }

    return xis


def test_xi_correlations(anscombes_xis):
    random.seed(2020)
    np.random.seed(2020)
    assert anscombes_xis["xi_1"].correlation == 0.2749999999999999
    assert anscombes_xis["xi_2"].correlation == 0.6
    assert anscombes_xis["xi_3"].correlation == 0.6190476190476191
    assert anscombes_xis["xi_4"].correlation == 0.1000000000000002


def test_p_val_asymptotic(anscombes_xis):
    random.seed(2020)
    np.random.seed(2020)
    # values taken from R code
    assert (
        anscombes_xis["xi_1"].pval_asymptotic(ties=False, nperm=1000)
        == 0.07841556446646347
    )
    assert (
        anscombes_xis["xi_2"].pval_asymptotic(ties=False, nperm=1000)
        == 0.0010040217037570187
    )
    assert (
        anscombes_xis["xi_3"].pval_asymptotic(ties=False, nperm=1000)
        == 0.04989192742513937
    )
    assert (
        anscombes_xis["xi_4"].pval_asymptotic(ties=False, nperm=1000)
        == 0.2599336349448975
    )



from pkg_resources import get_distribution, DistributionNotFound


__version__ = "0.1.0dev0"


__title__ = "xicor"
__description__ = "xi correlation for tied data"
__uri__ = "https://github.com/czbiohub/xicor"
__doc__ = __description__ + " <" + __uri__ + ">"

__author__ = "Phoenix Logan"
__email__ = "phoenix.logan@czbiohub.org"

__license__ = "MIT License"
__copyright__ = "Copyright (c) 2020 Chan Zuckerberg Biohub"


try:
    VERSION = get_distribution(__name__).version
except DistributionNotFound:  # pragma: no cover
    try:
        from .version import version as VERSION  # noqa
    except ImportError:  # pragma: no cover
        raise ImportError(
            "Failed to find (autogenerated) version.py. "
            "This might be because you are installing from GitHub's tarballs, "
            "use the PyPI ones."
        )



import pandas as pd
import numpy as np
import scipy.stats as ss
from collections import namedtuple


class XiPandas:
    def __init__(self, x, y):
        self.raw_x = pd.Series(x)
        self.raw_y = pd.Series(y)

    @property
    def sample_size(self):
        return len(self.raw_x)

    @property
    def data(self):
        CorrData = namedtuple("CorrData", ["x", "y"])
        both_not_null = self.raw_x.isnull() & self.raw_y.isnull()
        x = self.raw_x[~both_not_null]
        y = self.raw_y[~both_not_null]
        return CorrData(x, y)

    @property
    def x_ordered_rank(self):
        # PI is the rank vector for x, with ties broken at random
        # Not mine: source (https://stackoverflow.com/a/47430384/1628971)
        return self.data.x.sample(frac=1).rank(method="first").reindex_like(self.data.x)

    @property
    def y_rank_max(self):
        # f[i] is number of j s.t. y[j] <= y[i], divided by n.
        return self.data.y.rank(method="max") / self.sample_size

    @property
    def g(self):
        # g[i] is number of j s.t. y[j] >= y[i], divided by n.
        return (
            pd.Series([1 - i for i in self.data.y]).rank(method="max")
            / self.sample_size
        )

    @property
    def x_ordered(self):
        # order of the x's, ties broken at random.
        return np.argsort(self.x_ordered_rank)

    @property
    def x_rank_max_ordered(self):
        # Rearrange f according to ord.
        return [self.y_rank_max[i] for i in self.x_ordered]

    @property
    def mean_absolute(self):
        return (
            np.mean(
                np.abs(
                    [
                        x - y
                        for x, y in zip(
                            self.x_rank_max_ordered[0 : (self.sample_size - 1)],
                            self.x_rank_max_ordered[1 : self.sample_size],
                        )
                    ]
                )
            )
            * (self.sample_size - 1)
            / (2 * self.sample_size)
        )

    @property
    def inverse_g_mean(self):
        # C = np.mean(self.g*(1-self.g))
        return np.mean(self.g * (1 - self.g))

    @property
    def correlation(self):
        """xi correlation"""
        return 1 - self.mean_absolute / self.inverse_g_mean

    @classmethod
    def xi(cls, x, y):
        return cls(x, y)

    def pval_asymptotic(self, ties=False, nperm=1000, factor=True):
        # If there are no ties, return xi and theoretical P-value:

        if ties:
            return 1 - ss.norm.cdf(
                np.sqrt(self.sample_size) * self.correlation / np.sqrt(2 / 5)
            )

        # If there are ties, and the theoretical method is to be used for
        # calculation P-values:
        # The following steps calculate the theoretical variance
        # in the presence of ties:
        sorted_ordered_x_rank = sorted(self.x_rank_max_ordered)

        ind = [i + 1 for i in range(self.sample_size)]
        ind2 = [2 * self.sample_size - 2 * ind[i - 1] + 1 for i in ind]

        a = (
            np.mean([i * j * j for i, j in zip(ind2, sorted_ordered_x_rank)])
            / self.sample_size
        )

        c = (
            np.mean([i * j for i, j in zip(ind2, sorted_ordered_x_rank)])
            / self.sample_size
        )

        cq = np.cumsum(sorted_ordered_x_rank)

        m = [
            (i + (self.sample_size - j) * k) / self.sample_size
            for i, j, k in zip(cq, ind, sorted_ordered_x_rank)
        ]

        b = np.mean([np.square(i) for i in m])
        v = (a - 2 * b + np.square(c)) / np.square(self.inverse_g_mean)

        return 1 - ss.norm.cdf(
            np.sqrt(self.sample_size) * self.correlation / np.sqrt(v)
        )

    # def pval_permutation_test(self, nperm=1000):
    #     # If permutation test is to be used for calculating P-value:
    #     r = np.zeros((nperm,), dtype=float)
    #     for idx,val in enumerate(r):
    #     # x1 = runif(n, 0, 1)
    #         x1 = np.random.uniform(0, 1, self.sample_size)
    #         x1_xi = self.xi(x1, self.data.y).correlation
    #         r[idx] = x1_xi

    #     # Return xi and P-value based on permutation test:
    #     return (np.mean([ri for ri in r if ri > self.correlation]))



import setuptools

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("HISTORY.md") as history_file:
    history = history_file.read().replace(".. :changelog:", "")

with open("requirements.txt") as requirements_file:
    requirements = requirements_file.read()

test_requirements = ["pytest", "coverage", "pytest-cov"]


setuptools.setup(
    name="xicor",
    author="Phoenix Logan",
    author_email="phoenix.logan@czbiohub.org",
    maintainer="Pranathi Vemuri",
    maintainer_email="pranathi93.vemuri@gmail.com",
    description="xi correlation for tied data",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/czbiohub/xicor/",
    packages=setuptools.find_packages(
        exclude=["tests", "*.tests", "*.tests.*", "tests.*", "test_*"]
    ),
    include_package_data=True,
    install_requires=requirements,
    setup_requires=[
        "setuptools>=38.6.0",
        "setuptools_scm",
        "setuptools_scm_git_archive",
    ],
    license="MIT",
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    use_scm_version={"write_to": "xicor/version.py"},
    test_suite="tests",
    extras_require={"test": test_requirements},
)
